"""
Persistence model based on writing JSON documents.
"""

from __future__ import annotations

from contextlib import contextmanager
import copy
import json
from pathlib import Path
from typing import cast, TYPE_CHECKING
from typing_extensions import override

from fsspec import filesystem, AbstractFileSystem  # type: ignore
import numpy as np
from hpcflow.sdk.core import RUN_DIR_ARR_DTYPE, RUN_DIR_ARR_FILL
from hpcflow.sdk.core.errors import (
    MissingParameterData,
    MissingStoreEARError,
    MissingStoreElementError,
    MissingStoreElementIterationError,
)
from hpcflow.sdk.persistence.base import (
    PersistentStoreFeatures,
    PersistentStore,
    StoreEAR,
    StoreElement,
    StoreElementIter,
    StoreParameter,
    StoreTask,
    update_param_source_dict,
)
from hpcflow.sdk.submission.submission import JOBSCRIPT_SUBMIT_TIME_KEYS
from hpcflow.sdk.persistence.pending import CommitResourceMap
from hpcflow.sdk.persistence.store_resource import JSONFileStoreResource
from hpcflow.sdk.typing import DataIndex

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Mapping, Sequence
    from datetime import datetime
    from typing import Any, ClassVar, Literal
    from typing_extensions import Self
    from numpy.typing import NDArray
    from ..app import BaseApp
    from ..core.json_like import JSONed, JSONDocument
    from ..core.workflow import Workflow
    from ..typing import ParamSource
    from .types import (
        ElemMeta,
        IterMeta,
        LoopDescriptor,
        Metadata,
        RunMeta,
        StoreCreationInfo,
        TaskMeta,
        TemplateMeta,
    )


class JsonStoreTask(StoreTask["TaskMeta"]):
    """
    Persisted task that is serialized using JSON.
    """

    @override
    def encode(self) -> tuple[int, TaskMeta, dict[str, Any]]:
        """Prepare store task data for the persistent store."""
        assert self.task_template is not None
        wk_task: TaskMeta = {
            "id_": self.id_,
            "element_IDs": self.element_IDs,
            "index": self.index,
        }
        task = {"id_": self.id_, **self.task_template}
        return self.index, wk_task, task

    @override
    @classmethod
    def decode(cls, task_dat: TaskMeta) -> Self:
        """Initialise a `StoreTask` from store task data

        Note: the `task_template` is only needed for encoding because it is retrieved as
        part of the `WorkflowTemplate` so we don't need to load it when decoding.

        """
        return cls(is_pending=False, **task_dat)


class JsonStoreElement(StoreElement["ElemMeta", None]):
    """
    Persisted element that is serialized using JSON.
    """

    @override
    def encode(self, context: None) -> ElemMeta:
        """Prepare store element data for the persistent store."""
        dct = self.__dict__
        del dct["is_pending"]
        return cast("ElemMeta", dct)

    @override
    @classmethod
    def decode(cls, elem_dat: ElemMeta, context: None) -> Self:
        """Initialise a `JsonStoreElement` from store element data"""
        return cls(is_pending=False, **elem_dat)


class JsonStoreElementIter(StoreElementIter["IterMeta", None]):
    """
    Persisted element iteration that is serialized using JSON.
    """

    @override
    def encode(self, context: None) -> IterMeta:
        """Prepare store element iteration data for the persistent store."""
        dct = self.__dict__
        del dct["is_pending"]
        return cast("IterMeta", dct)

    @override
    @classmethod
    def decode(cls, iter_dat: IterMeta, context: None) -> Self:
        """Initialise a `JsonStoreElementIter` from persistent store element iteration data"""

        iter_dat = copy.deepcopy(iter_dat)  # to avoid mutating; can we avoid this?

        # cast JSON string keys to integers:
        if EAR_IDs := iter_dat["EAR_IDs"]:
            for act_idx in list(EAR_IDs):
                EAR_IDs[int(act_idx)] = EAR_IDs.pop(act_idx)

        return cls(is_pending=False, **cast("dict", iter_dat))


class JsonStoreEAR(StoreEAR["RunMeta", None]):
    """
    Persisted element action run that is serialized using JSON.
    """

    @override
    def encode(self, ts_fmt: str, context: None) -> RunMeta:
        """Prepare store EAR data for the persistent store."""
        return {
            "id_": self.id_,
            "elem_iter_ID": self.elem_iter_ID,
            "action_idx": self.action_idx,
            "commands_idx": self.commands_idx,
            "data_idx": self.data_idx,
            "submission_idx": self.submission_idx,
            "commands_file_ID": self.commands_file_ID,
            "success": self.success,
            "skip": self.skip,
            "start_time": self._encode_datetime(self.start_time, ts_fmt),
            "end_time": self._encode_datetime(self.end_time, ts_fmt),
            "snapshot_start": self.snapshot_start,
            "snapshot_end": self.snapshot_end,
            "exit_code": self.exit_code,
            "metadata": self.metadata,
            "run_hostname": self.run_hostname,
            "port_number": self.port_number,
        }

    @override
    @classmethod
    def decode(cls, EAR_dat: RunMeta, ts_fmt: str, context: None) -> Self:
        """Initialise a `JsonStoreEAR` from persistent store EAR data"""
        # don't want to mutate EAR_dat:
        EAR_dat = copy.deepcopy(EAR_dat)
        start_time = cls._decode_datetime(EAR_dat.pop("start_time"), ts_fmt)
        end_time = cls._decode_datetime(EAR_dat.pop("end_time"), ts_fmt)
        return cls(
            is_pending=False,
            **cast("dict", EAR_dat),
            start_time=start_time,
            end_time=end_time,
        )


class JSONPersistentStore(
    PersistentStore[
        JsonStoreTask,
        JsonStoreElement,
        JsonStoreElementIter,
        JsonStoreEAR,
        StoreParameter,
    ]
):
    """
    A store that writes JSON files for all its state serialization.
    """

    _name: ClassVar[str] = "json"
    _features: ClassVar[PersistentStoreFeatures] = PersistentStoreFeatures(
        create=True,
        edit=True,
        jobscript_parallelism=False,
        EAR_parallelism=False,
        schedulers=True,
        submission=True,
    )

    _meta_res: ClassVar[str] = "metadata"
    _params_res: ClassVar[str] = "parameters"
    _subs_res: ClassVar[str] = "submissions"
    _runs_res: ClassVar[str] = "runs"

    _res_file_names: ClassVar[Mapping[str, str]] = {
        _meta_res: "metadata.json",
        _params_res: "parameters.json",
        _subs_res: "submissions.json",
        _runs_res: "runs.json",
    }

    _res_map: ClassVar[CommitResourceMap] = CommitResourceMap(
        commit_tasks=(_meta_res,),
        commit_loops=(_meta_res,),
        commit_loop_num_iters=(_meta_res,),
        commit_loop_parents=(_meta_res,),
        commit_submissions=(_subs_res,),
        commit_at_submit_metadata=(_subs_res,),
        commit_js_metadata=(_subs_res,),
        commit_elem_IDs=(_meta_res,),
        commit_elements=(_meta_res,),
        commit_element_sets=(_meta_res,),
        commit_elem_iter_IDs=(_meta_res,),
        commit_elem_iters=(_meta_res,),
        commit_loop_indices=(_meta_res,),
        commit_elem_iter_EAR_IDs=(_meta_res,),
        commit_EARs_initialised=(_meta_res,),
        commit_EARs=(_runs_res,),
        commit_EAR_submission_indices=(_runs_res,),
        commit_EAR_skips=(_runs_res,),
        commit_EAR_starts=(_runs_res,),
        commit_EAR_ends=(_runs_res,),
        commit_template_components=(_meta_res,),
        commit_parameters=(_params_res,),
        commit_param_sources=(_params_res,),
        commit_set_run_dirs=(_runs_res,),
        commit_iter_data_idx=(_meta_res,),
        commit_run_data_idx=(_runs_res,),
    )

    @classmethod
    def _store_task_cls(cls) -> type[JsonStoreTask]:
        return JsonStoreTask

    @classmethod
    def _store_elem_cls(cls) -> type[JsonStoreElement]:
        return JsonStoreElement

    @classmethod
    def _store_iter_cls(cls) -> type[JsonStoreElementIter]:
        return JsonStoreElementIter

    @classmethod
    def _store_EAR_cls(cls) -> type[JsonStoreEAR]:
        return JsonStoreEAR

    @classmethod
    def _store_param_cls(cls) -> type[StoreParameter]:
        return StoreParameter

    def __init__(
        self, app, workflow: Workflow | None, path: Path, fs: AbstractFileSystem
    ):
        self._resources = {
            self._meta_res: self._get_store_resource(app, "metadata", path, fs),
            self._params_res: self._get_store_resource(app, "parameters", path, fs),
            self._subs_res: self._get_store_resource(app, "submissions", path, fs),
            self._runs_res: self._get_store_resource(app, "runs", path, fs),
        }
        super().__init__(app, workflow, path, fs)

        # store-specific cache data, assigned in `using_resource()` when
        # `_use_parameters_metadata_cache` is True, and set back to None when exiting the
        # `parameters_metadata_cache` context manager.
        self._parameters_file_dat: dict[str, dict[str, Any]] | None = None

    @contextmanager
    def cached_load(self) -> Iterator[None]:
        """Context manager to cache the metadata."""
        with self.using_resource("metadata", "read"):
            with self.using_resource("runs", "read"):
                yield

    @contextmanager
    def using_resource(
        self,
        res_label: Literal["metadata", "submissions", "parameters", "attrs", "runs"],
        action: str,
    ) -> Iterator[Any]:
        """Context manager for managing `StoreResource` objects associated with the store.

        Notes
        -----
        This overridden method facilitates easier use of the
        `JSONPersistentStore`-specific implementation of the `parameters_metadata_cache`,
        which in this case is just a copy of the `parameters.json` file data.

        """

        if (
            self._use_parameters_metadata_cache
            and res_label == "parameters"
            and action == "read"
        ):
            if not self._parameters_file_dat:
                with super().using_resource(
                    cast("Literal['parameters']", res_label), action
                ) as res:
                    self._parameters_file_dat = res
            yield self._parameters_file_dat

        else:
            with super().using_resource(res_label, action) as res:
                yield res

    @contextmanager
    def parameters_metadata_cache(self) -> Iterator[None]:
        """Context manager for using the parameters-metadata cache."""
        self._use_parameters_metadata_cache = True
        try:
            yield
        finally:
            self._use_parameters_metadata_cache = False
            self._parameters_file_dat = None  # clear cache data

    def remove_replaced_dir(self) -> None:
        """
        Remove the directory containing replaced workflow details.
        """
        with self.using_resource("metadata", "update") as md:
            if "replaced_workflow" in md:
                assert self.fs is not None
                self.remove_path(md["replaced_workflow"])
                self.logger.debug("removing temporarily renamed pre-existing workflow.")
                del md["replaced_workflow"]

    def reinstate_replaced_dir(self) -> None:
        """
        Reinstate the directory containing replaced workflow details.
        """
        with self.using_resource("metadata", "read") as md:
            if "replaced_workflow" in md:
                assert self.fs is not None
                self.logger.debug(
                    "reinstating temporarily renamed pre-existing workflow."
                )
                self.rename_path(md["replaced_workflow"], self.path)

    @classmethod
    def _get_store_resource(
        cls, app: BaseApp, name: str, path: str | Path, fs: AbstractFileSystem
    ) -> JSONFileStoreResource:
        return JSONFileStoreResource(
            app=app,
            name=name,
            path=path,
            fs=fs,
            filename=cls._res_file_names[name],
        )

    @classmethod
    def write_empty_workflow(
        cls,
        app: BaseApp,
        *,
        template_js: TemplateMeta,
        template_components_js: dict[str, Any],
        wk_path: str,
        fs: AbstractFileSystem,
        name: str,
        replaced_wk: str | None,
        creation_info: StoreCreationInfo,
        ts_fmt: str,
        ts_name_fmt: str,
    ) -> None:
        """
        Write an empty persistent workflow.
        """
        fs.mkdir(wk_path)
        submissions: list[None] = []
        parameters: dict[str, dict[None, None]] = {
            "data": {},
            "sources": {},
        }
        metadata: Metadata = {
            "name": name,
            "ts_fmt": ts_fmt,
            "ts_name_fmt": ts_name_fmt,
            "creation_info": creation_info,
            "template_components": template_components_js,
            "template": template_js,
            "tasks": [],
            "elements": [],
            "iters": [],
            "num_added_tasks": 0,
            "loops": [],
        }
        runs: dict[str, list] = {
            "runs": [],
            "run_dirs": [],
        }
        if replaced_wk:
            metadata["replaced_workflow"] = replaced_wk

        cls._get_store_resource(app, "metadata", wk_path, fs)._dump(metadata)
        cls._get_store_resource(app, "parameters", wk_path, fs)._dump(parameters)
        cls._get_store_resource(app, "submissions", wk_path, fs)._dump(submissions)
        cls._get_store_resource(app, "runs", wk_path, fs)._dump(runs)

    def _append_tasks(self, tasks: Iterable[StoreTask]):
        with self.using_resource("metadata", action="update") as md:
            assert "tasks" in md and "template" in md and "num_added_tasks" in md
            for task in tasks:
                idx, wk_task_i, task_i = task.encode()
                md["tasks"].insert(idx, cast("TaskMeta", wk_task_i))
                md["template"]["tasks"].insert(idx, task_i)
                md["num_added_tasks"] += 1

    def _append_loops(self, loops: dict[int, LoopDescriptor]):
        with self.using_resource("metadata", action="update") as md:
            assert "loops" in md and "template" in md
            for _, loop in loops.items():
                md["loops"].append(
                    {
                        "num_added_iterations": loop["num_added_iterations"],
                        "iterable_parameters": loop["iterable_parameters"],
                        "output_parameters": loop["output_parameters"],
                        "parents": loop["parents"],
                    }
                )
                md["template"]["loops"].append(loop["loop_template"])

    def _append_submissions(self, subs: dict[int, Mapping[str, JSONed]]):
        with self.using_resource("submissions", action="update") as subs_res:
            subs_res.extend(subs.values())

    def _append_task_element_IDs(self, task_ID: int, elem_IDs: list[int]):
        with self.using_resource("metadata", action="update") as md:
            assert "tasks" in md
            md["tasks"][task_ID]["element_IDs"].extend(elem_IDs)

    def _append_elements(self, elems: Sequence[JsonStoreElement]):
        with self.using_resource("metadata", action="update") as md:
            assert "elements" in md
            md["elements"].extend(elem.encode(None) for elem in elems)

    def _append_element_sets(self, task_id: int, es_js: Sequence[Mapping]):
        task_idx = self._get_task_id_to_idx_map()[task_id]
        with self.using_resource("metadata", "update") as md:
            assert "template" in md
            md["template"]["tasks"][task_idx]["element_sets"].extend(es_js)

    def _append_elem_iter_IDs(self, elem_ID: int, iter_IDs: Iterable[int]):
        with self.using_resource("metadata", action="update") as md:
            assert "elements" in md
            md["elements"][elem_ID]["iteration_IDs"].extend(iter_IDs)

    def _append_elem_iters(self, iters: Sequence[JsonStoreElementIter]):
        with self.using_resource("metadata", action="update") as md:
            assert "iters" in md
            md["iters"].extend(it.encode(None) for it in iters)

    def _append_elem_iter_EAR_IDs(
        self, iter_ID: int, act_idx: int, EAR_IDs: Sequence[int]
    ):
        with self.using_resource("metadata", action="update") as md:
            assert "iters" in md
            md["iters"][iter_ID].setdefault("EAR_IDs", {}).setdefault(act_idx, []).extend(
                EAR_IDs
            )

    def _update_elem_iter_EARs_initialised(self, iter_ID: int):
        with self.using_resource("metadata", action="update") as md:
            assert "iters" in md
            md["iters"][iter_ID]["EARs_initialised"] = True

    def _update_at_submit_metadata(self, at_submit_metadata: dict[int, dict[str, Any]]):
        with self.using_resource("submissions", action="update") as subs_res:
            for sub_idx, metadata_i in at_submit_metadata.items():
                sub = subs_res[sub_idx]
                assert isinstance(sub, dict)
                for dt_str, parts_j in metadata_i["submission_parts"].items():
                    sub["at_submit_metadata"]["submission_parts"][dt_str] = parts_j

    def _update_loop_index(self, loop_indices: dict[int, dict[str, int]]):
        with self.using_resource("metadata", action="update") as md:
            assert "iters" in md
            for iter_ID, loop_idx in loop_indices.items():
                md["iters"][iter_ID]["loop_idx"].update(loop_idx)

    def _update_loop_num_iters(self, index: int, num_iters: list[list[list[int] | int]]):
        with self.using_resource("metadata", action="update") as md:
            assert "loops" in md
            md["loops"][index]["num_added_iterations"] = num_iters

    def _update_loop_parents(self, index: int, parents: list[str]):
        with self.using_resource("metadata", action="update") as md:
            assert "loops" in md
            md["loops"][index]["parents"] = parents

    def _update_iter_data_indices(self, iter_data_indices: dict[int, DataIndex]):
        with self.using_resource("metadata", action="update") as md:
            assert "iters" in md
            for iter_ID, dat_idx in iter_data_indices.items():
                md["iters"][iter_ID]["data_idx"].update(dat_idx)

    def _update_run_data_indices(self, run_data_indices: dict[int, DataIndex]):
        with self.using_resource("runs", action="update") as md:
            assert "runs" in md
            for run_ID, dat_idx in run_data_indices.items():
                md["runs"][run_ID]["data_idx"].update(dat_idx)

    def _append_EARs(self, EARs: Sequence[JsonStoreEAR]):
        with self.using_resource("runs", action="update") as md:
            assert "runs" in md
            assert "run_dirs" in md
            md["runs"].extend(i.encode(self.ts_fmt, None) for i in EARs)
            md["run_dirs"].extend([None] * len(EARs))

    def _set_run_dirs(self, run_dir_arr: np.ndarray, run_idx: np.ndarray):
        with self.using_resource("runs", action="update") as md:
            assert "run_dirs" in md
            dirs_lst = md["run_dirs"]
            for idx, r_idx in enumerate(run_idx):
                dirs_lst[r_idx] = run_dir_arr[idx].item()
            md["run_dirs"] = dirs_lst

    def _update_EAR_submission_data(self, sub_data: Mapping[int, tuple[int, int | None]]):
        with self.using_resource("runs", action="update") as md:
            assert "runs" in md
            for EAR_ID_i, (sub_idx_i, cmd_file_ID) in sub_data.items():
                md["runs"][EAR_ID_i]["submission_idx"] = sub_idx_i
                md["runs"][EAR_ID_i]["commands_file_ID"] = cmd_file_ID

    def _update_EAR_start(
        self,
        run_starts: dict[int, tuple[datetime, dict[str, Any] | None, str, int | None]],
    ):
        with self.using_resource("runs", action="update") as md:
            assert "runs" in md
            for run_id, (s_time, s_snap, s_hn, port_number) in run_starts.items():
                md["runs"][run_id]["start_time"] = s_time.strftime(self.ts_fmt)
                md["runs"][run_id]["snapshot_start"] = s_snap
                md["runs"][run_id]["run_hostname"] = s_hn
                md["runs"][run_id]["port_number"] = port_number

    def _update_EAR_end(
        self, run_ends: dict[int, tuple[datetime, dict[str, Any] | None, int, bool]]
    ):
        with self.using_resource("runs", action="update") as md:
            assert "runs" in md
            for run_id, (e_time, e_snap, ext_code, success) in run_ends.items():
                md["runs"][run_id]["end_time"] = e_time.strftime(self.ts_fmt)
                md["runs"][run_id]["snapshot_end"] = e_snap
                md["runs"][run_id]["exit_code"] = ext_code
                md["runs"][run_id]["success"] = success

    def _update_EAR_skip(self, skips: dict[int, int]):
        with self.using_resource("runs", action="update") as md:
            assert "runs" in md
            for run_ID, reason in skips.items():
                md["runs"][run_ID]["skip"] = reason

    def _update_js_metadata(self, js_meta: dict[int, dict[int, dict[str, Any]]]):
        with self.using_resource("submissions", action="update") as sub_res:
            for sub_idx, all_js_md in js_meta.items():
                sub = cast("dict[str, list[dict[str, Any]]]", sub_res[sub_idx])
                for js_idx, js_meta_i in all_js_md.items():
                    self.logger.info(
                        f"updating jobscript metadata for (sub={sub_idx}, js={js_idx}): "
                        f"{js_meta_i!r}."
                    )
                    _at_submit_md = {
                        k: js_meta_i.pop(k)
                        for k in JOBSCRIPT_SUBMIT_TIME_KEYS
                        if k in js_meta_i
                    }
                    sub["jobscripts"][js_idx].update(**js_meta_i)
                    sub["jobscripts"][js_idx]["at_submit_metadata"].update(
                        **_at_submit_md
                    )

    def _append_parameters(self, params: Sequence[StoreParameter]):
        with self.using_resource("parameters", "update") as params_u:
            for param_i in params:
                params_u["data"][str(param_i.id_)] = param_i.encode()
                params_u["sources"][str(param_i.id_)] = param_i.source

    def _set_parameter_values(self, set_parameters: dict[int, tuple[Any, bool]]):
        """Set multiple unset persistent parameters."""
        param_objs = self._get_persistent_parameters(set_parameters)
        with self.using_resource("parameters", "update") as params:
            for param_id, (value, is_file) in set_parameters.items():
                param_i = param_objs[param_id]
                if is_file:
                    param_i = param_i.set_file(value)
                else:
                    param_i = param_i.set_data(value)
                params["data"][str(param_id)] = param_i.encode()

    def _update_parameter_sources(self, sources: Mapping[int, ParamSource]):
        """Update the sources of multiple persistent parameters."""
        param_objs = self._get_persistent_parameters(sources)
        with self.using_resource("parameters", "update") as params:
            # no need to update data array:
            for p_id, src_i in sources.items():
                param_i = param_objs[p_id]
                new_src_i = update_param_source_dict(param_i.source, src_i)
                params["sources"][str(p_id)] = new_src_i

    def _update_template_components(self, tc: dict[str, Any]):
        with self.using_resource("metadata", "update") as md:
            md["template_components"] = tc

    def _get_num_persistent_tasks(self) -> int:
        """Get the number of persistent tasks."""
        if self.use_cache and self.num_tasks_cache is not None:
            num = self.num_tasks_cache
        else:
            with self.using_resource("metadata", action="read") as md:
                assert "tasks" in md
                num = len(md["tasks"])
        if self.use_cache and self.num_tasks_cache is None:
            self.num_tasks_cache = num
        return num

    def _get_num_persistent_loops(self) -> int:
        """Get the number of persistent loops."""
        with self.using_resource("metadata", action="read") as md:
            assert "loops" in md
            return len(md["loops"])

    def _get_num_persistent_submissions(self) -> int:
        """Get the number of persistent submissions."""
        with self.using_resource("submissions", "read") as subs_res:
            return len(subs_res)

    def _get_num_persistent_elements(self) -> int:
        """Get the number of persistent elements."""
        with self.using_resource("metadata", action="read") as md:
            assert "elements" in md
            return len(md["elements"])

    def _get_num_persistent_elem_iters(self) -> int:
        """Get the number of persistent element iterations."""
        with self.using_resource("metadata", action="read") as md:
            assert "iters" in md
            return len(md["iters"])

    def _get_num_persistent_EARs(self) -> int:
        """Get the number of persistent EARs."""
        if self.use_cache and self.num_EARs_cache is not None:
            num = self.num_EARs_cache
        else:
            with self.using_resource("runs", action="read") as md:
                assert "runs" in md
                num = len(md["runs"])
        if self.use_cache and self.num_EARs_cache is None:
            self.num_EARs_cache = num
        return num

    def _get_num_persistent_parameters(self) -> int:
        if self.use_cache and self.num_params_cache is not None:
            num = self.num_params_cache
        else:
            with self.using_resource("parameters", "read") as params:
                assert "data" in params
                num = len(params["data"])
        if self.use_cache and self.num_params_cache is None:
            self.num_params_cache = num
        return num

    def _get_num_persistent_added_tasks(self) -> int:
        with self.using_resource("metadata", "read") as md:
            assert "num_added_tasks" in md
            return md["num_added_tasks"]

    @classmethod
    def make_test_store_from_spec(
        cls,
        app: BaseApp,
        spec,
        dir=None,
        path="test_store.json",
        overwrite=False,
        ts_fmt="%d/%m/%Y, %H:%M:%S",  # FIXME: use the right default timestamp format
    ):
        """Generate an store for testing purposes."""

        tasks_, elems, elem_iters, EARs = super().prepare_test_store_from_spec(spec)

        path_ = Path(path).resolve()
        tasks = [JsonStoreTask(**task_info).encode() for task_info in tasks_]
        elements_ = [JsonStoreElement(**elem_info).encode(None) for elem_info in elems]
        elem_iters_ = [
            JsonStoreElementIter(**it_info).encode(None) for it_info in elem_iters
        ]
        EARs_ = [JsonStoreEAR(**ear_info).encode(ts_fmt, None) for ear_info in EARs]

        persistent_data = {
            "tasks": tasks,
            "elements": elements_,
            "iters": elem_iters_,
            "runs": EARs_,
        }

        path_ = Path(dir or "", path_)
        with path_.open("wt") as fp:
            json.dump(persistent_data, fp, indent=2)

        return cls(app=app, workflow=None, path=path_, fs=filesystem("file"))

    def _get_persistent_template_components(self) -> dict[str, Any]:
        with self.using_resource("metadata", "read") as md:
            assert "template_components" in md
            return md["template_components"]

    def _get_persistent_template(self) -> dict[str, JSONed]:
        with self.using_resource("metadata", "read") as md:
            assert "template" in md
            return cast("dict[str, JSONed]", md["template"])

    def _get_persistent_tasks(self, id_lst: Iterable[int]) -> dict[int, JsonStoreTask]:
        tasks, id_lst = self._get_cached_persistent_tasks(id_lst)
        if id_lst:
            with self.using_resource("metadata", action="read") as md:
                assert "tasks" in md
                new_tasks = {
                    i["id_"]: JsonStoreTask.decode({**i, "index": idx})
                    for idx, i in enumerate(cast("Sequence[TaskMeta]", md["tasks"]))
                    if id_lst is None or i["id_"] in id_lst
                }
                self.task_cache.update(new_tasks)
                tasks.update(new_tasks)
        return tasks

    def _get_persistent_loops(
        self, id_lst: Iterable[int] | None = None
    ) -> dict[int, LoopDescriptor]:
        with self.using_resource("metadata", "read") as md:
            assert "loops" in md
            return {
                idx: cast("LoopDescriptor", i)
                for idx, i in enumerate(md["loops"])
                if id_lst is None or idx in id_lst
            }

    def _get_persistent_submissions(
        self, id_lst: Iterable[int] | None = None
    ) -> dict[int, Mapping[str, JSONed]]:
        with self.using_resource("submissions", "read") as sub_res:
            subs_dat = copy.deepcopy(
                {
                    idx: i
                    for idx, i in enumerate(sub_res)
                    if id_lst is None or idx in id_lst
                }
            )
            # cast jobscript `task_elements` keys:
            for sub in subs_dat.values():
                js: dict[str, Any]
                assert isinstance(sub, dict)
                for js in sub["jobscripts"]:
                    blk: dict[str, Any]
                    assert isinstance(js, dict)
                    for blk in js["blocks"]:
                        for key in list(te := blk["task_elements"]):
                            te[int(key)] = te.pop(key)

        return subs_dat

    def _get_persistent_elements(
        self, id_lst: Iterable[int]
    ) -> dict[int, JsonStoreElement]:
        elems, id_lst_ = self._get_cached_persistent_elements(id_lst)
        if id_lst_:
            # could convert `id_lst` to e.g. slices if more efficient for a given store
            with self.using_resource("metadata", action="read") as md:
                try:
                    if "elements" not in md:
                        raise KeyError
                    elem_dat = {id_: md["elements"][id_] for id_ in id_lst_}
                except KeyError:
                    raise MissingStoreElementError(id_lst_)
                new_elems = {
                    k: JsonStoreElement.decode(v, None) for k, v in elem_dat.items()
                }
                self.element_cache.update(new_elems)
                elems.update(new_elems)
        return elems

    def _get_persistent_element_iters(
        self, id_lst: Iterable[int]
    ) -> dict[int, JsonStoreElementIter]:
        iters, id_lst_ = self._get_cached_persistent_element_iters(id_lst)
        if id_lst_:
            with self.using_resource("metadata", action="read") as md:
                try:
                    if "iters" not in md:
                        raise KeyError
                    iter_dat = {id_: md["iters"][id_] for id_ in id_lst_}
                except KeyError:
                    raise MissingStoreElementIterationError(id_lst_)
                new_iters = {
                    k: JsonStoreElementIter.decode(v, None) for k, v in iter_dat.items()
                }
                self.element_iter_cache.update(new_iters)
                iters.update(new_iters)
        return iters

    def _get_persistent_EARs(self, id_lst: Iterable[int]) -> dict[int, JsonStoreEAR]:
        runs, id_lst_ = self._get_cached_persistent_EARs(id_lst)
        if id_lst_:
            with self.using_resource("runs", action="read") as md:
                try:
                    if "runs" not in md:
                        raise KeyError
                    EAR_dat = {id_: md["runs"][id_] for id_ in id_lst_}
                except KeyError:
                    raise MissingStoreEARError(id_lst_)
                new_runs = {
                    k: JsonStoreEAR.decode(v, self.ts_fmt, None)
                    for k, v in EAR_dat.items()
                }
                self.EAR_cache.update(new_runs)
                runs.update(new_runs)
        return runs

    def _get_persistent_parameters(
        self, id_lst: Iterable[int], **kwargs
    ) -> Mapping[int, StoreParameter]:
        params, id_lst_ = self._get_cached_persistent_parameters(id_lst)
        if id_lst_:
            with self.using_resource("parameters", "read") as params_:
                try:
                    param_dat = {id_: params_["data"][str(id_)] for id_ in id_lst_}
                    src_dat = {id_: params_["sources"][str(id_)] for id_ in id_lst_}
                except KeyError:
                    raise MissingParameterData(id_lst_)

            new_params = {
                k: StoreParameter.decode(id_=k, data=v, source=src_dat[k])
                for k, v in param_dat.items()
            }
            self.parameter_cache.update(new_params)
            params.update(new_params)
        return params

    def _get_persistent_param_sources(
        self, id_lst: Iterable[int]
    ) -> dict[int, ParamSource]:
        sources, id_lst_ = self._get_cached_persistent_param_sources(id_lst)
        if id_lst_:
            with self.using_resource("parameters", "read") as params:
                try:
                    new_sources = {id_: params["sources"][str(id_)] for id_ in id_lst_}
                except KeyError:
                    raise MissingParameterData(id_lst_)
            self.param_sources_cache.update(new_sources)
            sources.update(new_sources)
        return sources

    def _get_persistent_parameter_set_status(
        self, id_lst: Iterable[int]
    ) -> dict[int, bool]:
        with self.using_resource("parameters", "read") as params:
            try:
                param_dat = {id_: params["data"][str(id_)] for id_ in id_lst}
            except KeyError:
                raise MissingParameterData(id_lst)
        return {k: v is not None for k, v in param_dat.items()}

    def _get_persistent_parameter_IDs(self) -> list[int]:
        with self.using_resource("parameters", "read") as params:
            return [int(i) for i in params["data"]]

    def get_ts_fmt(self) -> str:
        """
        Get the format for timestamps.
        """
        with self.using_resource("metadata", action="read") as md:
            assert "ts_fmt" in md
            return md["ts_fmt"]

    def get_ts_name_fmt(self) -> str:
        """
        Get the format for timestamps to use in names.
        """
        with self.using_resource("metadata", action="read") as md:
            assert "ts_name_fmt" in md
            return md["ts_name_fmt"]

    def get_creation_info(self) -> StoreCreationInfo:
        """
        Get information about the creation of the workflow.
        """
        with self.using_resource("metadata", action="read") as md:
            assert "creation_info" in md
            return copy.deepcopy(md["creation_info"])

    def get_name(self) -> str:
        """
        Get the name of the workflow.
        """
        with self.using_resource("metadata", action="read") as md:
            assert "name" in md
            return md["name"]

    def zip(
        self,
        path: str = ".",
        log: str | None = None,
        overwrite=False,
        include_execute=False,
        include_rechunk_backups=False,
    ) -> str:
        raise TypeError("unsupported operation: zipping-json")

    def unzip(self, path: str = ".", log: str | None = None) -> str:
        raise TypeError("unsupported operation: unzipping-json")

    def rechunk_parameter_base(
        self,
        chunk_size: int | None = None,
        backup: bool = True,
        status: bool = True,
    ) -> Any:
        raise TypeError("unsupported operation: rechunk-json")

    def rechunk_runs(
        self,
        chunk_size: int | None = None,
        backup: bool = True,
        status: bool = True,
    ) -> Any:
        raise TypeError("unsupported operation: rechunk-json")

    def get_dirs_array(self) -> NDArray:
        """
        Retrieve the run directories array.
        """
        with self.using_resource("runs", action="read") as md:
            dirs_lst = md["run_dirs"]
            dirs_arr = np.zeros(len(dirs_lst), dtype=RUN_DIR_ARR_DTYPE)
            dirs_arr[:] = RUN_DIR_ARR_FILL
            for idx, i in enumerate(dirs_lst):
                if i is not None:
                    dirs_arr[idx] = tuple(i)
        return dirs_arr



"""
Types to support the core SDK.
"""

from __future__ import annotations
from typing import Any, Literal, Protocol, TYPE_CHECKING
from typing_extensions import NotRequired, TypeAlias, TypedDict

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from datetime import datetime, timedelta
    import numpy as np
    from valida.conditions import ConditionLike  # type: ignore
    from .actions import ActionScope
    from .command_files import FileSpec
    from .enums import ParallelMode, ParameterPropagationMode
    from .object_list import ResourceList
    from .parameters import (
        InputSource,
        InputValue,
        Parameter,
        ResourceSpec,
    )
    from .task import InputStatus
    from ..persistence.types import ParamSource


class ParameterDependence(TypedDict):
    """
    Dependency descriptor for a parameter.
    """

    #: The input file writers that can use the parameter.
    input_file_writers: list[FileSpec]
    #: The commands that can use the parameter.
    commands: list[int]


class ScriptData(TypedDict, total=False):
    """
    Descriptor for data relating to a script.
    """

    #: The format of the data.
    format: str
    #: Whether the data is required for all iterations.
    all_iterations: NotRequired[bool]


class JobscriptSubmissionFailureArgs(TypedDict):
    """
    Arguments that can be expanded to create a
    :class:`JobscriptSubmissionFailure`.
    """

    #: The command that was submitted.
    submit_cmd: list[str]
    #: The jobscript index.
    js_idx: int
    #: The jobscript path.
    js_path: str
    #: Where to write stdout.
    stdout: NotRequired[str]
    #: Where to write stderr.
    stderr: NotRequired[str]
    #: The exception from the exec of the subprocess.
    subprocess_exc: NotRequired[Exception]
    #: The exception from parsing the job ID.
    job_ID_parse_exc: NotRequired[Exception]


class ElementDescriptor(TypedDict):
    """
    Descriptor for elements.
    """

    #: The statuses of inputs.
    input_statuses: Mapping[str, InputStatus]
    #: The sources of inputs.
    input_sources: Mapping[str, InputSource]
    #: The insertion ID.
    task_insert_ID: int


class _DependentDescriptor(TypedDict):
    #: The names of groups of dependents.
    group_names: tuple[str, ...]


class DependentDescriptor(_DependentDescriptor, total=False):
    """
    Descriptor for dependents.
    """


class IterableParam(TypedDict):
    """
    The type of the descriptor for an iterable parameter.
    """

    #: Identifier for the input task supplying the parameter.
    input_task: int
    #: Identifiers for the output tasks consuming the parameter.
    output_tasks: list[int]


#: Type of an address.
Address: TypeAlias = "list[int | float | str]"
#: Type of something numeric.
Numeric: TypeAlias = "int | float | np.number"


class LabelInfo(TypedDict):
    """
    Information about a label.
    """

    #: The label propagation mode, if known.
    propagation_mode: NotRequired[ParameterPropagationMode]
    #: The group containing the label, if known.
    group: NotRequired[str]
    #: The default value for the label, if known.
    default_value: NotRequired[InputValue]


class LabellingDescriptor(TypedDict):
    """
    Descriptor for a labelling.
    """

    #: The type with the label.
    labelled_type: str
    #: The propagation mode for the label.
    propagation_mode: ParameterPropagationMode
    #: The group containing the label.
    group: str | None
    #: The default value for the label, if known.
    default_value: NotRequired[InputValue]


class ResourceSpecArgs(TypedDict):
    """
    Supported keyword arguments for a ResourceSpec.
    """

    #: Which scope does this apply to.
    scope: NotRequired[ActionScope | str]
    #: Which scratch space to use.
    scratch: NotRequired[str]
    #: Which parallel mode to use.
    parallel_mode: NotRequired[str | ParallelMode]
    #: How many cores to request.
    num_cores: NotRequired[int]
    #: How many cores per compute node to request.
    num_cores_per_node: NotRequired[int]
    #: How many threads to request.
    num_threads: NotRequired[int]
    #: How many compute nodes to request.
    num_nodes: NotRequired[int]
    #: Which scheduler to use.
    scheduler: NotRequired[str]
    #: Which system shell to use.
    shell: NotRequired[str]
    #: Whether to use array jobs.
    use_job_array: NotRequired[bool]
    #: If using array jobs, up to how many items should be in the job array.
    max_array_items: NotRequired[int]
    #: How long to run for.
    time_limit: NotRequired[str | timedelta]
    #: Additional arguments to pass to the scheduler.
    scheduler_args: NotRequired[dict[str, Any]]
    #: Additional arguments to pass to the shell.
    shell_args: NotRequired[dict[str, Any]]
    #: Which OS to use.
    os_name: NotRequired[str]
    #: Which execution environments to use.
    environments: NotRequired[dict[str, dict[str, Any]]]
    #: Which SGE parallel environment to request.
    SGE_parallel_env: NotRequired[str]
    #: Which SLURM partition to request.
    SLURM_partition: NotRequired[str]
    #: How many SLURM tasks to request.
    SLURM_num_tasks: NotRequired[str]
    #: How many SLURM tasks per compute node to request.
    SLURM_num_tasks_per_node: NotRequired[str]
    #: How many compute nodes to request.
    SLURM_num_nodes: NotRequired[str]
    #: How many CPU cores to ask for per SLURM task.
    SLURM_num_cpus_per_task: NotRequired[str]


# Used in declaration of Resources below
_R: TypeAlias = "ResourceSpec | ResourceSpecArgs | dict"
#: The type of things we can normalise to a :py:class:`ResourceList`.
Resources: TypeAlias = "_R | ResourceList | None | Sequence[_R]"


class SchemaInputKwargs(TypedDict):
    """
    Just used when deep copying `SchemaInput`.
    """

    #: The parameter.
    parameter: Parameter | str
    #: Whether this is multiple.
    multiple: bool
    #: The labels.
    labels: dict[str, LabelInfo] | None
    #: The number or proportional of permitted unset parameter data found when reolving
    #: this input from upstream outputs.
    allow_failed_dependencies: int | float | bool | None


class RuleArgs(TypedDict):
    """
    The keyword arguments that may be used to create a Rule.
    """

    #: If present, check this attribute exists.
    check_exists: NotRequired[str]
    #: If present, check this attribute does *not* exist.
    check_missing: NotRequired[str]
    #: Where to look up the attribute to check.
    #: If not present, determined by context.
    path: NotRequired[str]
    #: If present, a general condition to check (or kwargs used to generate one).
    condition: NotRequired[dict[str, Any] | ConditionLike]
    #: If present, a cast to apply prior to running the general check.
    cast: NotRequired[str]
    #: Optional descriptive text.
    doc: NotRequired[str]


class ActParameterDependence(TypedDict):
    """
    Action parameter dependency descriptor.
    """

    #: The input file writers that produce the parameter.
    input_file_writers: list[tuple[int, FileSpec]]
    #: The commands that produce the parameter.
    commands: list[tuple[int, int]]


#: A relevant path when applying an update.
RelevantPath: TypeAlias = "ParentPath | UpdatePath | SiblingPath"


class RepeatsDescriptor(TypedDict):
    """
    Descriptor for repeats.
    """

    #: Name of the repeat.
    name: str
    #: The repeat count.
    number: int
    #: The nesting order. Normally an integer; non-integer values have special meanings.
    nesting_order: float


class MultiplicityDescriptor(TypedDict):
    """
    Descriptor for multiplicities.
    """

    #: The size of the multiplicity.
    multiplicity: int
    #: The nesting order. Normally an integer; non-integer values have special meanings.
    nesting_order: float
    #: The path to the multiplicity.
    path: str


class ParentPath(TypedDict):
    """
    A `RelevantPath` that is a path to a parent.
    """

    #: Type ID.
    type: Literal["parent"]
    relative_path: Sequence[str]


class UpdatePath(TypedDict):
    """
    A `RelevantPath` that is a path to an update.
    """

    #: Type ID.
    type: Literal["update"]
    update_path: Sequence[str]


class SiblingPath(TypedDict):
    """
    A `RelevantPath` that is a path to a sibling.
    """

    #: Type ID.
    type: Literal["sibling"]


class RelevantData(TypedDict):
    """
    Data relevant to performing an update.
    """

    #: The data to set.
    data: list[Any] | Any
    #: Which method to use for handling the data, if any.
    value_class_method: list[str | None] | str | None
    #: Whether the value is set.
    is_set: bool | list[bool]
    #: Whether the value is multiple.
    is_multi: bool


class CreationInfo(TypedDict):
    """
    Descriptor for creation information about a workflow.
    """

    #: Description of information about the application.
    app_info: dict[str, Any]
    #: When the workflow was created.
    create_time: datetime
    #: Unique identifier for the workflow.
    id: str


class WorkflowTemplateTaskData(TypedDict):
    """
    Descriptor for information about tasks described in a workflow template.
    """

    #: The schema, if known.
    schema: NotRequired[Any | list[Any]]
    #: The element sets, if known.
    element_sets: NotRequired[list[WorkflowTemplateElementSetData]]
    #: The output labels, if known.
    output_labels: NotRequired[list[str]]


class WorkflowTemplateElementSetData(TypedDict):
    """
    Descriptor for element set data within a workflow template parametrisation.
    """

    #: Inputs to the set of elements.
    inputs: NotRequired[list[dict[str, Any]]]
    #: Input files to the set of elements.
    input_files: NotRequired[list[dict[str, Any]]]
    #: Description of how to repeat the set of elements.
    repeats: NotRequired[int | list[RepeatsDescriptor]]
    #: Groupings in the set of elements.
    groups: NotRequired[list[dict[str, Any]]]
    #: Resources to use for the set of elements.
    resources: NotRequired[dict[str, Any]]
    #: Input value sequences to parameterise over.
    sequences: NotRequired[list[dict[str, Any]]]
    #: Input value multi-path sequences to parameterise over.
    multi_path_sequences: NotRequired[list[dict[str, Any]]]
    #: Input source descriptors.
    input_sources: NotRequired[dict[str, list]]
    #: How to handle nesting of iterations.
    nesting_order: NotRequired[dict[str, float]]
    #: Which environment preset to use.
    env_preset: NotRequired[str]
    #: Environment descriptors to use.
    environments: NotRequired[dict[str, dict[str, Any]]]
    #: List of global element iteration indices from which inputs for
    #: the new elements associated with this element set may be sourced.
    #: If ``None``, all iterations are valid.
    sourceable_elem_iters: NotRequired[list[int]]
    #: Whether to allow sources to come from distinct element sub-sets.
    allow_non_coincident_task_sources: NotRequired[bool]
    #: Whether this initialisation is the first for this data (i.e. not a
    #: reconstruction from persistent workflow data), in which case, we merge
    #: ``environments`` into ``resources`` using the "any" scope, and merge any multi-
    #: path sequences into the sequences list.
    is_creation: NotRequired[bool]


class Pending(TypedDict):
    """
    Pending update information. Internal use only.
    """

    #: Template components to update.
    template_components: dict[str, list[int]]
    #: Tasks to update.
    tasks: list[int]
    #: Loops to update.
    loops: list[int]
    #: Submissions to update.
    submissions: list[int]


class AbstractFileSystem(Protocol):
    """
    Type constraints for an abstract file system.
    """

    # Because a dependency is not fully typed...
    def exists(self, path: str) -> bool:
        """Test if a path points to a file or directory that exists."""

    def rename(self, from_: str, to: str, *, recursive: bool = False) -> None:
        """Rename a file or directory."""

    def rm(self, path: str, *, recursive: bool = False) -> None:
        """Delete a file or directory."""

    def glob(self, pattern: str) -> list[str]:
        """List files in a directory that match a pattern."""


class ResourcePersistingWorkflow(Protocol):
    """
    An object to pass to :py:meth:`ResourceSpec.make_persistent` that handles
    persisting resources.
    """

    def _add_parameter_data(self, data: Any, source: ParamSource) -> int:
        ...

    def check_parameters_exist(self, id_lst: int | list[int]) -> bool:
        """
        Check if all the parameters exist.
        """


BlockActionKey: TypeAlias = "tuple[int | str, int | str, int | str]"
"""
The type of indices that locate an action within a submission. The indices represent,
respectively, the jobscript index, the jobscript-block index, and the block-action index.
Usually, these are integers, but in the case of strings, they will correspond to shell
environment variables.
"""



"""
Operating system information discovery helpers.
"""

from __future__ import annotations
from collections.abc import Mapping
import platform
import re
import subprocess
from typing import Final

_DEFAULT_LINUX_RELEASE_FILE: Final = "/etc/os-release"


def get_OS_info() -> Mapping[str, str]:
    """
    Get basic operating system version info.
    """
    uname = platform.uname()
    return {
        "OS_name": uname.system,
        "OS_version": uname.version,
        "OS_release": uname.release,
    }


def get_OS_info_windows() -> Mapping[str, str]:
    """
    Get operating system version info: Windows version.
    """
    return get_OS_info()


def get_OS_info_POSIX(
    WSL_executable: list[str] | None = None,
    use_py: bool = True,
    linux_release_file: str | None = None,
) -> Mapping[str, str]:
    """
    Get operating system version info: POSIX version.

    Parameters
    ----------
    WSL_executable:
        Executable to run subprocess calls via WSL on Windows.
    use_py:
        If True, use the :py:func:`platform.uname` Python function to get the OS
        information. Otherwise use subprocess to call ``uname``. We set this to False
        when getting OS info in WSL on Windows, since we need to call the WSL executable.
    linux_release_file:
        If on Linux, record the name and version fields from this file.
    """

    def try_subprocess_call(*args: str) -> str:
        exc = None
        command = [*WSL_exe, *args]
        try:
            proc = subprocess.run(
                args=command,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )
        except Exception as err:
            exc = err

        if proc.returncode or exc:
            raise RuntimeError(
                f"Failed to get POSIX OS info. Command was: {command!r}. Subprocess "
                f"exception was: {exc!r}. Stderr was: {proc.stderr!r}."
            )
        return proc.stdout

    WSL_exe = WSL_executable or []
    out: dict[str, str] = {}
    if use_py:
        out.update(**get_OS_info())
    else:
        OS_name = try_subprocess_call("uname", "-s").strip()
        OS_release = try_subprocess_call("uname", "-r").strip()
        OS_version = try_subprocess_call("uname", "-v").strip()

        out["OS_name"] = OS_name
        out["OS_release"] = OS_release
        out["OS_version"] = OS_version

    if out["OS_name"] == "Linux":
        # get linux distribution name and version:
        linux_release_file = linux_release_file or _DEFAULT_LINUX_RELEASE_FILE
        release_out = try_subprocess_call("cat", linux_release_file)

        name_match = _NAME_RE.search(release_out)
        if not name_match:
            raise RuntimeError(
                f"Failed to get Linux distribution name from file `{linux_release_file}`."
            )
        lin_name: str = name_match[1]

        version_match = _VERSION_RE.search(release_out)
        if not version_match:
            raise RuntimeError(
                f"Failed to get Linux distribution version from file "
                f"`{linux_release_file}`."
            )
        lin_version: str = version_match[1]

        out["linux_release_file"] = linux_release_file
        out["linux_distribution_name"] = lin_name
        out["linux_distribution_version"] = lin_version

    return out


_NAME_RE: Final = re.compile(r"^NAME=\"(.*)\"", flags=re.MULTILINE)
_VERSION_RE: Final = re.compile(r"^VERSION=\"(.*)\"", flags=re.MULTILINE)



"""
Shell models based on Microsoft PowerShell.
"""

from __future__ import annotations
import subprocess
from textwrap import dedent
from typing import TYPE_CHECKING
from typing_extensions import override
from hpcflow.sdk.typing import hydrate
from hpcflow.sdk.submission.shells.base import Shell
from hpcflow.sdk.submission.shells.os_version import get_OS_info_windows

if TYPE_CHECKING:
    from typing import ClassVar
    from .base import VersionInfo


@hydrate
class WindowsPowerShell(Shell):
    """Class to represent using PowerShell on Windows to generate and submit a jobscript."""

    # TODO: add snippets that can be used in demo task schemas?

    #: Default for executable name.
    DEFAULT_EXE: ClassVar[str] = "powershell.exe"

    #: File extension for jobscripts.
    JS_EXT: ClassVar[str] = ".ps1"
    #: Basic indent.
    JS_INDENT: ClassVar[str] = "    "
    #: Indent for environment setup.
    JS_ENV_SETUP_INDENT: ClassVar[str] = 2 * JS_INDENT
    #: Template for the jobscript shebang line.
    JS_SHEBANG: ClassVar[str] = ""
    #: Template for the jobscript functions file.
    JS_FUNCS: ClassVar[str] = dedent(
        """\
        function {workflow_app_alias} {{
            & {{
        {env_setup}{app_invoc} `
                    --with-config log_file_path "$env:{app_caps}_LOG_PATH" `
                    --config-dir "{config_dir}" `
                    --config-key "{config_invoc_key}" `
                    $args
            }} @args
        }}

        function get_nth_line($file, $line) {{
            Get-Content $file | Select-Object -Skip $line -First 1
        }}
    """
    )
    #: Template for the common part of the jobscript header.
    JS_HEADER: ClassVar[str] = dedent(
        """\
        $ErrorActionPreference = 'Stop'

        function JoinMultiPath {{
            $numArgs = $args.Length
            $path = $args[0]
            for ($i = 1; $i -lt $numArgs; $i++) {{
                $path = Join-Path $path $args[$i]
            }}
            return $path
        }}        

        $WK_PATH = $(Get-Location)
        $WK_PATH_ARG = $WK_PATH
        $SUB_IDX = {sub_idx}
        $JS_IDX = {js_idx}

        $SUB_DIR = JoinMultiPath $WK_PATH artifacts submissions $SUB_IDX
        $JS_FUNCS_PATH = JoinMultiPath $SUB_DIR {jobscript_functions_dir} {jobscript_functions_name}
        . $JS_FUNCS_PATH

        $EAR_ID_FILE = JoinMultiPath $SUB_DIR {run_IDs_file_dir} {run_IDs_file_name}
        $SUB_TMP_DIR = Join-Path $SUB_DIR {tmp_dir_name}
        $SUB_LOG_DIR = Join-Path $SUB_DIR {log_dir_name}
        $SUB_STD_DIR = Join-Path $SUB_DIR {app_std_dir_name}
        $SUB_SCRIPTS_DIR = Join-Path $SUB_DIR {scripts_dir_name}        

        $env:{app_caps}_WK_PATH = $WK_PATH
        $env:{app_caps}_WK_PATH_ARG = $WK_PATH_ARG
        $env:{app_caps}_SUB_IDX = {sub_idx}
        $env:{app_caps}_SUB_SCRIPTS_DIR = $SUB_SCRIPTS_DIR
        $env:{app_caps}_SUB_TMP_DIR = $SUB_TMP_DIR
        $env:{app_caps}_SUB_LOG_DIR = $SUB_LOG_DIR
        $env:{app_caps}_SUB_STD_DIR = $SUB_STD_DIR                
        $env:{app_caps}_LOG_PATH = Join-Path $SUB_LOG_DIR "js_$JS_IDX.log"
        $env:{app_caps}_JS_FUNCS_PATH = $JS_FUNCS_PATH
        $env:{app_caps}_JS_IDX = {js_idx}
        $env:{app_caps}_RUN_ID_FILE = $EAR_ID_FILE
    """
    )
    #: Template for the jobscript header when directly executed.
    JS_DIRECT_HEADER: ClassVar[str] = dedent(
        """\
        {shebang}
        {header}
        {wait_command}
    """
    )
    #: Template for enabling writing of the app log.
    JS_RUN_LOG_PATH_ENABLE: ClassVar[str] = 'Join-Path $SUB_LOG_DIR "{run_log_file_name}"'
    #: Template for disabling writing of the app log.
    JS_RUN_LOG_PATH_DISABLE: ClassVar[str] = '" "'
    #: Template for the run execution command.
    JS_RUN_CMD: ClassVar[str] = (
        "{workflow_app_alias} internal workflow $WK_PATH execute-run "
        "$SUB_IDX $JS_IDX $block_idx $block_act_idx $EAR_ID\n"
    )
    #: Template for the execution command for multiple combined runs.
    JS_RUN_CMD_COMBINED: ClassVar[str] = (
        "{workflow_app_alias} internal workflow $WK_PATH execute-combined-runs "
        "$SUB_IDX $JS_IDX\n"
    )
    #: Template for setting up run environment variables and executing the run.
    JS_RUN: ClassVar[str] = dedent(
        """\
        $EAR_ID = ($elem_EAR_IDs -split "{EAR_files_delimiter}")[$block_act_idx]
        if ($EAR_ID -eq -1) {{
            continue
        }}

        $env:{app_caps}_RUN_ID = $EAR_ID
        $env:{app_caps}_RUN_LOG_PATH = {run_log_enable_disable}
        $env:{app_caps}_LOG_PATH = $env:{app_caps}_RUN_LOG_PATH
        $env:{app_caps}_RUN_STD_PATH = Join-Path $SUB_STD_DIR "$env:{app_caps}_RUN_ID.txt"
        $env:{app_caps}_BLOCK_ACT_IDX = $block_act_idx            

        Set-Location $SUB_TMP_DIR
        
        {run_cmd}
        """
    )
    #: Template for the action-run processing loop in a jobscript.
    JS_ACT_MULTI: ClassVar[str] = dedent(
        """\
        for ($block_act_idx = 0; $block_act_idx -lt {num_actions}; $block_act_idx += 1) {{        
        {run_block}
        }}
        """
    )
    #: Template for the single-action-run execution in a jobscript.
    JS_ACT_SINGLE: ClassVar[str] = dedent(
        """\
        $block_act_idx = 0        
        {run_block}
        """
    )
    #: Template for setting up environment variables and running one or more action-runs.
    JS_MAIN: ClassVar[str] = dedent(
        """\
        $block_elem_idx = ($JS_elem_idx - {block_start_elem_idx})
        $elem_EAR_IDs = get_nth_line $EAR_ID_FILE $JS_elem_idx
        $env:{app_caps}_JS_ELEM_IDX = $JS_elem_idx
        $env:{app_caps}_BLOCK_ELEM_IDX = $block_elem_idx

        {action}
    """
    )
    #: Template for a jobscript-block header.
    JS_BLOCK_HEADER: ClassVar[str] = dedent(  # for single-block jobscripts only
        """\
        $block_idx = 0
        $env:{app_caps}_BLOCK_IDX = 0
        """
    )
    #: Template for single-element execution.
    JS_ELEMENT_SINGLE: ClassVar[str] = dedent(
        """\
        $JS_elem_idx = {block_start_elem_idx}
        {main}
    """
    )
    #: Template for the element processing loop in a jobscript.
    JS_ELEMENT_MULTI_LOOP: ClassVar[str] = dedent(
        """\
        for ($JS_elem_idx = {block_start_elem_idx}; $JS_elem_idx -lt ({block_start_elem_idx} + {num_elements}); $JS_elem_idx += 1) {{            
        {main}
        }}
    """
    )
    #: Template for the jobscript block loop in a jobscript.
    JS_BLOCK_LOOP: ClassVar[str] = dedent(
        """\
        $num_elements = {num_elements}
        $num_actions = {num_actions}
        $block_start_elem_idx = 0
        for ($block_idx = 0; $block_idx -lt {num_blocks}; $block_idx += 1 ) {{
            $env:{app_caps}_BLOCK_IDX = $block_idx
        {element_loop}
            $block_start_elem_idx += $num_elements[$block_idx]
        }}
    """
    )
    #: Template for the jobscript footer.
    JS_FOOTER: ClassVar[str] = dedent(
        """\
        Set-Location $WK_PATH
    """
    )

    def get_direct_submit_command(self, js_path: str) -> list[str]:
        """Get the command for submitting a non-scheduled jobscript."""
        return [*self.executable, "-File", js_path]

    def get_command_file_launch_command(self, cmd_file_path: str) -> list[str]:
        """Get the command for launching the commands file for a given run."""
        # note the "-File" argument is required for the correct exit code to be recorded.
        return [*self.executable, "-File", cmd_file_path]

    @override
    def get_version_info(self, exclude_os: bool = False) -> VersionInfo:
        """Get powershell version information.

        Parameters
        ----------
        exclude_os
            If True, exclude operating system information.

        """

        proc = subprocess.run(
            args=self.executable + ["-Command", "$PSVersionTable.PSVersion.ToString()"],
            stdout=subprocess.PIPE,
            text=True,
        )
        if proc.returncode == 0:
            PS_version = proc.stdout.strip()
        else:
            raise RuntimeError("Failed to parse PowerShell version information.")

        osinfo = {} if exclude_os else get_OS_info_windows()
        return {
            "shell_name": "powershell",
            "shell_executable": self.executable,
            "shell_version": PS_version,
            **osinfo,
        }

    @staticmethod
    def process_app_invoc_executable(app_invoc_exe: str) -> str:
        if " " in app_invoc_exe:
            # use call operator and single-quote the executable path:
            app_invoc_exe = f"& '{app_invoc_exe}'"
        return app_invoc_exe

    @override
    def format_env_var_get(self, var: str) -> str:
        """
        Format retrieval of a shell environment variable.
        """
        return f"$env:{var}"

    @override
    def format_array(self, lst: list) -> str:
        """
        Format construction of a shell array.
        """
        return "@(" + ", ".join(str(i) for i in lst) + ")"

    @override
    def format_array_get_item(self, arr_name: str, index: int | str) -> str:
        """
        Format retrieval of a shell array item at a specified index.
        """
        return f"${arr_name}[{index}]"

    @override
    def format_stream_assignment(self, shell_var_name: str, command: str) -> str:
        """
        Produce code to assign the output of the command to a shell variable.
        """
        return f"${shell_var_name} = {command}"

    @override
    def format_source_functions_file(self, app_name: str, commands: str) -> str:
        """
        Format sourcing (i.e. invocation) of the jobscript functions file.
        """
        app_caps = app_name.upper()
        out = dedent(
            """\
            . $env:{app_name}_JS_FUNCS_PATH

            """
        ).format(app_name=app_caps)

        var_strings = (
            f"{app_caps}_WK_PATH",
            f"{app_caps}_SUB_SCRIPTS_DIR",
            f"{app_caps}_JS_IDX",
            f"{app_caps}_BLOCK_IDX",
            f"{app_caps}_BLOCK_ACT_IDX",
            f"{app_caps}_RUN_ID",
            f"{app_caps}_RUN_STD_PATH",
            f"{app_caps}_RUN_SCRIPT_NAME",
            f"{app_caps}_RUN_SCRIPT_NAME_NO_EXT",
            f"{app_caps}_RUN_SCRIPT_DIR",
            f"{app_caps}_RUN_SCRIPT_PATH",
        )
        add = False
        for i in var_strings:
            if i in commands:
                add = True
                out += f"${i} = $env:{i}\n"

        if add:
            out += "\n"

        return out

    @override
    def format_commands_file(self, app_name: str, commands: str) -> str:
        """
        Format the commands file.
        """
        return (
            self.format_source_functions_file(app_name, commands)
            + commands
            + "\nexit $LASTEXITCODE\n"
        )

    @override
    def format_save_parameter(
        self,
        workflow_app_alias: str,
        param_name: str,
        shell_var_name: str,
        cmd_idx: int,
        stderr: bool,
        app_name: str,
    ) -> str:
        """
        Produce code to save a parameter's value into the workflow persistent store.
        """
        # TODO: quote shell_var_name as well? e.g. if it's a white-space delimited list?
        #   and test.
        stderr_str = " --stderr" if stderr else ""
        app_caps = app_name.upper()
        return (
            f'{workflow_app_alias} --std-stream "${app_caps}_RUN_STD_PATH" '
            f'internal workflow "${app_caps}_WK_PATH" save-parameter {stderr_str}'
            f'"--" {param_name} ${shell_var_name} ${app_caps}_RUN_ID {cmd_idx}'
            f"\n"
        )



import os
from pathlib import Path
import pytest
from hpcflow.app import app as hf


@pytest.mark.integration
def test_zarr_metadata_file_modification_times_many_jobscripts(null_config, tmp_path):
    """Test that root group attributes are modified first, then individual jobscript
    at-submit-metadata chunk files, then the submission at-submit-metadata group
    attributes."""

    num_js = 30
    t1 = hf.Task(
        schema=hf.task_schemas.test_t1_conditional_OS,
        inputs={"p1": 100},
        sequences=[
            hf.ValueSequence(
                path="resources.any.resources_id", values=list(range(num_js))
            )
        ],
    )
    wk = hf.Workflow.from_template_data(
        template_name="test_zarr_metadata_attrs_modified_times",
        path=tmp_path,
        tasks=[t1],
        store="zarr",
    )
    wk.submit(add_to_known=False, status=False, cancel=True)

    mtime_meta_group = Path(wk.path).joinpath(".zattrs").stat().st_mtime
    mtime_mid_jobscript_chunk = (
        wk._store._get_jobscripts_at_submit_metadata_arr_path(0)
        .joinpath(str(int(num_js / 2)))
        .stat()
        .st_mtime
    )
    mtime_submission_group = (
        wk._store._get_submission_metadata_group_path(0)
        .joinpath(".zattrs")
        .stat()
        .st_mtime
    )
    assert mtime_meta_group < mtime_mid_jobscript_chunk < mtime_submission_group


@pytest.mark.integration
def test_json_metadata_file_modification_times_many_jobscripts(null_config, tmp_path):
    """Test that the metadata.json file is modified first, then the submissions.json
    file."""

    num_js = 30
    t1 = hf.Task(
        schema=hf.task_schemas.test_t1_conditional_OS,
        inputs={"p1": 100},
        sequences=[
            hf.ValueSequence(
                path="resources.any.resources_id", values=list(range(num_js))
            )
        ],
    )
    wk = hf.Workflow.from_template_data(
        template_name="test_zarr_metadata_attrs_modified_times",
        path=tmp_path,
        tasks=[t1],
        store="json",
    )
    wk.submit(add_to_known=False, status=False, cancel=True)

    mtime_meta = Path(wk.path).joinpath("metadata.json").stat().st_mtime
    mtime_subs = Path(wk.path).joinpath("submissions.json").stat().st_mtime
    assert mtime_meta < mtime_subs


@pytest.mark.integration
def test_subission_start_end_times_equal_to_first_and_last_jobscript_start_end_times(
    null_config, tmp_path
):
    num_js = 2
    t1 = hf.Task(
        schema=hf.task_schemas.test_t1_conditional_OS,
        inputs={"p1": 100},
        sequences=[
            hf.ValueSequence(
                path="resources.any.resources_id", values=list(range(num_js))
            )
        ],
    )
    wk = hf.Workflow.from_template_data(
        template_name="test_subission_start_end_times",
        path=tmp_path,
        tasks=[t1],
    )
    wk.submit(wait=True, add_to_known=False, status=False)

    sub = wk.submissions[0]
    jobscripts = sub.jobscripts

    assert len(jobscripts) == num_js

    # submission has two jobscripts, so start time should be start time of first jobscript:
    assert sub.start_time == jobscripts[0].start_time

    # ...and end time should be end time of second jobscript:
    assert sub.end_time == jobscripts[1].end_time


@pytest.mark.integration
def test_multiple_jobscript_functions_files(null_config, tmp_path):
    if os.name == "nt":
        shell_exes = ["powershell.exe", "pwsh.exe", "pwsh.exe"]
    else:
        shell_exes = ["/bin/bash", "bash", "bash"]
    t1 = hf.Task(
        schema=hf.task_schemas.test_t1_conditional_OS,
        inputs={"p1": 100},
        sequences=[
            hf.ValueSequence(
                path="resources.any.shell_args.executable",
                values=shell_exes,
            )
        ],
    )
    wk = hf.Workflow.from_template_data(
        template_name="test_multi_js_funcs_files",
        path=tmp_path,
        tasks=[t1],
        store="json",
    )
    wk.submit(add_to_known=True, status=False, cancel=True)

    sub_js = wk.submissions[0].jobscripts
    assert len(sub_js) == 2

    funcs_0 = sub_js[0].jobscript_functions_path
    funcs_1 = sub_js[1].jobscript_functions_path

    assert funcs_0.is_file()
    assert funcs_1.is_file()
    assert funcs_0 != funcs_1



"""
Serialization and deserialization mechanism intended to map between a complex
graph of objects and either JSON or YAML.
"""
from __future__ import annotations

from collections import defaultdict
from collections.abc import Container, Sequence, Mapping
import copy
from dataclasses import dataclass
import enum
from types import SimpleNamespace
from typing import overload, Protocol, cast, runtime_checkable, TYPE_CHECKING
from typing_extensions import final, override

from hpcflow.sdk.core.app_aware import AppAware
from hpcflow.sdk.typing import hydrate
from hpcflow.sdk import app, get_SDK_logger
from hpcflow.sdk.core.utils import get_md5_hash
from hpcflow.sdk.core.validation import get_schema
from hpcflow.sdk.core.errors import ToJSONLikeChildReferenceError

if TYPE_CHECKING:
    from typing import Any, ClassVar, Literal
    from typing_extensions import Self, TypeAlias, TypeIs
    from ..app import BaseApp
    from .object_list import ObjectList

_BasicJsonTypes: TypeAlias = "int | float | str | None"
_WriteStructure: TypeAlias = (
    "list[JSONable] | tuple[JSONable, ...] | set[JSONable] | dict[str, JSONable]"
)
JSONDocument: TypeAlias = "Sequence[JSONed] | Mapping[str, JSONed]"
JSONable: TypeAlias = "_WriteStructure | enum.Enum | BaseJSONLike | _BasicJsonTypes"
JSONed: TypeAlias = "JSONDocument | _BasicJsonTypes"

if TYPE_CHECKING:
    _ChildType: TypeAlias = "type[enum.Enum | JSONLike]"
    _JSONDeserState: TypeAlias = "dict[str, dict[str, JSONed]] | None"


#: Primitive types supported by the serialization mechanism.
PRIMITIVES = (
    int,
    float,
    str,
    type(None),
)

_SDK_logger = get_SDK_logger(__name__)


@runtime_checkable
class _AltConstructFromJson(Protocol):
    @classmethod
    def _json_like_constructor(cls, json_like: Mapping[str, JSONed]) -> Self:
        pass


def _is_base_json_like(value: JSONable) -> TypeIs[BaseJSONLike]:
    return value is not None and hasattr(value, "to_json_like")


_MAX_DEPTH = 50


@overload
def to_json_like(
    obj: int,
    shared_data: _JSONDeserState = None,
    parent_refs: dict | None = None,
    path: list | None = None,
) -> tuple[int, _JSONDeserState]:
    ...


@overload
def to_json_like(
    obj: float,
    shared_data: _JSONDeserState = None,
    parent_refs: dict | None = None,
    path: list | None = None,
) -> tuple[float, _JSONDeserState]:
    ...


@overload
def to_json_like(
    obj: str,
    shared_data: _JSONDeserState = None,
    parent_refs: dict | None = None,
    path: list | None = None,
) -> tuple[str, _JSONDeserState]:
    ...


@overload
def to_json_like(
    obj: None,
    shared_data: _JSONDeserState = None,
    parent_refs: dict | None = None,
    path: list | None = None,
) -> tuple[None, _JSONDeserState]:
    ...


@overload
def to_json_like(
    obj: enum.Enum,
    shared_data: _JSONDeserState = None,
    parent_refs: dict | None = None,
    path: list | None = None,
) -> tuple[str, _JSONDeserState]:
    ...


@overload
def to_json_like(
    obj: list[JSONable],
    shared_data: _JSONDeserState = None,
    parent_refs: dict | None = None,
    path: list | None = None,
) -> tuple[Sequence[JSONed], _JSONDeserState]:
    ...


@overload
def to_json_like(
    obj: tuple[JSONable, ...],
    shared_data: _JSONDeserState = None,
    parent_refs: dict | None = None,
    path: list | None = None,
) -> tuple[Sequence[JSONed], _JSONDeserState]:
    ...


@overload
def to_json_like(
    obj: set[JSONable],
    shared_data: _JSONDeserState = None,
    parent_refs: dict | None = None,
    path: list | None = None,
) -> tuple[Sequence[JSONed], _JSONDeserState]:
    ...


@overload
def to_json_like(
    obj: dict[str, JSONable],
    shared_data: _JSONDeserState = None,
    parent_refs: dict | None = None,
    path: list | None = None,
) -> tuple[Mapping[str, JSONed], _JSONDeserState]:
    ...


@overload
def to_json_like(
    obj: BaseJSONLike,
    shared_data: _JSONDeserState = None,
    parent_refs: dict | None = None,
    path: list | None = None,
) -> tuple[Mapping[str, JSONed], _JSONDeserState]:
    ...


def to_json_like(
    obj: JSONable,
    shared_data: _JSONDeserState = None,
    parent_refs: dict | None = None,
    path: list | None = None,
):
    """
    Convert the object to a JSON-like basic value tree.
    Such trees are trivial to serialize as JSON or YAML.
    """
    path = path or []

    if len(path) > _MAX_DEPTH:
        raise RuntimeError(f"I'm in too deep! Path is: {path}")

    if isinstance(obj, (list, tuple, set)):
        out_list: list[JSONed] = []
        for idx, item in enumerate(obj):
            if _is_base_json_like(item):
                new_item, shared_data = item.to_json_like(
                    shared_data=shared_data,
                    exclude=frozenset((parent_refs or {}).values()),
                    path=[*path, idx],
                )
                out_list.append(new_item)
            else:
                new_std_item, shared_data = to_json_like(
                    item, shared_data=shared_data, path=[*path, idx]
                )
                out_list.append(new_std_item)
        if isinstance(obj, tuple):
            return tuple(out_list), shared_data
        elif isinstance(obj, set):
            return set(out_list), shared_data
        else:
            return out_list, shared_data

    elif isinstance(obj, dict):
        out_map: dict[str, JSONed] = {}
        for dct_key, dct_val in obj.items():
            if _is_base_json_like(dct_val):
                try:
                    out_map[dct_key], shared_data = dct_val.to_json_like(
                        shared_data=shared_data,
                        exclude={(parent_refs or {}).get(dct_key)},
                        path=[*path, dct_key],
                    )
                except ToJSONLikeChildReferenceError:
                    continue
            else:
                out_map[dct_key], shared_data = to_json_like(
                    dct_val,
                    shared_data=shared_data,
                    parent_refs=parent_refs,
                    path=[*path, dct_key],
                )
        return out_map, shared_data

    elif isinstance(obj, PRIMITIVES):
        return obj, shared_data

    elif isinstance(obj, enum.Enum):
        return obj.name, shared_data

    else:
        return obj.to_json_like(shared_data=shared_data, path=path)


@dataclass
class ChildObjectSpec:
    """
    Used to describe what the child structure of an class is so that the generic
    deserializer can build the structure.
    """

    #: The name of the attribute.
    name: str
    #: The name of the class (or class of members of a list) used to deserialize the
    #: attribute.
    class_name: str | None = None
    #: The class (or class of members of a list) used to deserialize the
    #: attribute.
    class_obj: type[enum.Enum | BaseJSONLike] | None = None
    # TODO: no need for class_obj/class_name if shared data?
    #: The name of the key used in the JSON document, if different from the attribute
    #: name.
    json_like_name: str | None = None
    #: If true, the attribute is really a list of instances,
    #: or a dictionary if :attr:`dict_key_attr` is set.
    is_multiple: bool = False
    #: If set, the name of an attribute of the object to use as a dictionary key.
    #: Requires that :attr:`is_multiple` be set as well.
    dict_key_attr: str | None = None
    #: If set, the name of an attribute of the object to use as a dictionary value.
    #: If not set but :attr:`dict_key_attr` is set, the whole object is the value.
    #: Requires that :attr:`dict_key_attr` be set as well.
    dict_val_attr: str | None = None
    #: If set, the attribute of the child object that contains a reference to its parent.
    parent_ref: str | None = None
    # TODO: do parent refs make sense when from shared? Prob not.
    #: If true, enables special handling where there can be only one child descriptor
    #: for a containing class.
    is_single_attribute: bool = False
    #: If true, the object is an enum member and should use special serialization rules.
    is_enum: bool = False
    #: If true, the child object is a dict, whose values are of the specified class.
    #: The dict structure will remain.
    is_dict_values: bool = False
    #: If true, values that are not lists are cast to lists and multiple child objects
    #: are instantiated for each dict value.
    is_dict_values_ensure_list: bool = False
    #: What key to look values up under in the shared data cache.
    #: If unspecified, the shared data cache is ignored.
    shared_data_name: str | None = None
    #: What attribute provides the value of the key into the shared data cache.
    #: If unspecified, a hash of the object dictionary is used.
    #: Ignored if :py:attr:`~.shared_data_name` is unspecified.
    shared_data_primary_key: str | None = None
    # shared_data_secondary_keys: tuple[str, ...] | None = None # TODO: what's the point?

    def __post_init__(self) -> None:
        if self.class_name and self.class_obj:
            raise ValueError("Specify at most one of `class_name` and `class_obj`.")

        if self.dict_key_attr and not isinstance(self.dict_key_attr, str):
            raise TypeError(
                "`dict_key_attr` must be of type `str`, but has type "
                f"{type(self.dict_key_attr)} with value {self.dict_key_attr}."
            )  # TODO: test raise
        if self.dict_val_attr:
            if not self.dict_key_attr:
                raise ValueError(
                    "If `dict_val_attr` is specified, `dict_key_attr` must be specified."
                )  # TODO: test raise
            if not isinstance(self.dict_val_attr, str):
                raise TypeError(
                    "`dict_val_attr` must be of type `str`, but has type "
                    f"{type(self.dict_val_attr)} with value {self.dict_val_attr}."
                )  # TODO: test raise
        if not self.is_multiple and self.dict_key_attr:
            raise ValueError(
                "If `dict_key_attr` is specified, `is_multiple` must be set to True."
            )
        if not self.is_multiple and self.is_dict_values:
            raise ValueError(
                "If `is_dict_values` is specified, `is_multiple` must be set to True."
            )
        if self.is_dict_values_ensure_list and not self.is_dict_values:
            raise ValueError(
                "If `is_dict_values_ensure_list` is specified, `is_dict_values` must be "
                "set to True."
            )
        if self.parent_ref and not isinstance(self.parent_ref, str):
            raise TypeError(
                "`parent_ref` must be of type `str`, but has type "
                f"{type(self.parent_ref)} with value {self.parent_ref}."
            )  # TODO: test raise

        self.json_like_name = self.json_like_name or self.name


@hydrate
class BaseJSONLike:
    """
    An object that has a serialization as JSON or YAML.

    Parameters
    ----------
    _class_namespace : namespace
        Namespace whose attributes include the class definitions that might be
        referenced (and so require instantiation) in child objects.
    _shared_data_namespace : namespace
        Namespace whose attributes include the shared data that might be referenced
        in child objects.
    """

    _child_objects: ClassVar[Sequence[ChildObjectSpec]] = ()
    _validation_schema: ClassVar[str | None] = None

    __class_namespace: ClassVar[dict[str, Any] | SimpleNamespace | BaseApp | None] = None
    _hash_value: str | None

    @overload
    @classmethod
    def _set_class_namespace(
        cls, value: SimpleNamespace, is_dict: Literal[False] = False
    ) -> None:
        ...

    @overload
    @classmethod
    def _set_class_namespace(cls, value: dict[str, Any], is_dict: Literal[True]) -> None:
        ...

    @classmethod
    def _set_class_namespace(
        cls, value: dict[str, Any] | SimpleNamespace, is_dict=False
    ) -> None:
        cls.__class_namespace = value

    @classmethod
    def _class_namespace(cls) -> dict[str, Any] | SimpleNamespace | BaseApp:
        if (ns := cls.__class_namespace) is None:
            raise ValueError(f"`{cls.__name__}` `class_namespace` must be set!")
        return ns

    @classmethod
    def __get_child_class(cls, child_spec: ChildObjectSpec) -> _ChildType | None:
        if child_spec.class_obj:
            return cast("_ChildType", child_spec.class_obj)
        elif child_spec.class_name:
            ns = cls._class_namespace()
            if isinstance(ns, dict):
                return ns[child_spec.class_name]
            else:
                return getattr(ns, child_spec.class_name)
        else:
            return None

    @classmethod
    def _get_default_shared_data(cls) -> Mapping[str, ObjectList[JSONable]]:
        return {}

    @overload
    @classmethod
    def from_json_like(
        cls,
        json_like: str,
        shared_data: Mapping[str, ObjectList[JSONable]] | None = None,
    ) -> Self | None:
        ...

    @overload
    @classmethod
    def from_json_like(
        cls,
        json_like: Sequence[Mapping[str, JSONed]] | Mapping[str, JSONed],
        shared_data: Mapping[str, ObjectList[JSONable]] | None = None,
    ) -> Self:
        ...

    @overload
    @classmethod
    def from_json_like(
        cls,
        json_like: None,
        shared_data: Mapping[str, ObjectList[JSONable]] | None = None,
    ) -> None:
        ...

    @classmethod
    def from_json_like(
        cls,
        json_like: str | Mapping[str, JSONed] | Sequence[Mapping[str, JSONed]] | None,
        shared_data: Mapping[str, ObjectList[JSONable]] | None = None,
    ) -> Self | None:
        """
        Make an instance of this class from JSON (or YAML) data.

        Parameters
        ----------
        json_like:
            The data to deserialise.
        shared_data:
            Shared context data.

        Returns
        -------
            The deserialised object.
        """
        shared_data = shared_data or cls._get_default_shared_data()
        if isinstance(json_like, str):
            json_like = cls._parse_from_string(json_like)
        if json_like is None:
            # e.g. optional attributes # TODO: is this still needed?
            return None
        return cls._from_json_like(copy.deepcopy(json_like), shared_data)

    @classmethod
    def _parse_from_string(cls, string: str) -> dict[str, str] | None:
        raise TypeError(f"unparseable {cls}: '{string}'")

    @staticmethod
    def __listify(v: JSONed, spec: ChildObjectSpec) -> list[JSONed]:
        if spec.is_dict_values_ensure_list and isinstance(v, list):
            return v
        return [v]

    @classmethod
    def __remap_child_seq(
        cls, spec: ChildObjectSpec, json_like: JSONed
    ) -> tuple[list[JSONed], dict[str, list[int]]]:
        if not spec.is_multiple:
            return [json_like], {}
        elif isinstance(json_like, list):
            return json_like, {}
        elif not isinstance(json_like, dict):
            raise TypeError(
                f"Child object {spec.name} of {cls.__name__!r} must be a list or "
                f"dict, but is of type {type(json_like)} with value {json_like!r}."
            )

        multi_chd_objs: list[JSONed] = []

        if spec.is_dict_values:
            # (if is_dict_values) indices into multi_chd_objs that enable reconstruction
            # of the source dict:
            is_dict_values_idx: dict[str, list[int]] = defaultdict(list)

            # keep as a dict
            for k, v in json_like.items():
                for item in cls.__listify(v, spec):
                    is_dict_values_idx[k].append(len(multi_chd_objs))
                    multi_chd_objs.append(item)
            return multi_chd_objs, is_dict_values_idx

        # want to cast to a list
        if not spec.dict_key_attr:
            raise ValueError(
                f"{cls.__name__!r}: must specify a `dict_key_attr` for child "
                f"object spec {spec.name!r}."
            )

        for k, v in json_like.items():
            all_attrs: dict[str, JSONed] = {spec.dict_key_attr: k}
            if spec.dict_val_attr:
                all_attrs[spec.dict_val_attr] = v
            elif isinstance(v, dict):
                all_attrs.update(v)
            else:
                raise TypeError(
                    f"Value for key {k!r} must be a dict representing "
                    f"attributes of the {spec.name!r} child object "
                    f"(parent: {cls.__name__!r}). If it instead "
                    f"represents a single attribute, set the "
                    f"`dict_val_attr` of the child object spec."
                )
            multi_chd_objs.append(all_attrs)

        return multi_chd_objs, {}

    @classmethod
    def __inflate_enum(cls, chd_cls: type[enum.Enum], multi_chd_objs: list[JSONed]):
        out: list[JSONable] = []
        for item in multi_chd_objs:
            if item is None:
                out.append(None)
            elif not isinstance(item, str):
                raise ValueError(
                    f"Enumeration {chd_cls!r} has no name {item!r}. Available"
                    f" names are: {chd_cls._member_names_!r}."
                )
            else:
                try:
                    out.append(getattr(chd_cls, item.upper()))
                except AttributeError:
                    raise ValueError(
                        f"Enumeration {chd_cls!r} has no name {item!r}. Available"
                        f" names are: {chd_cls._member_names_!r}."
                    )
        return out

    @classmethod
    def _from_json_like(
        cls,
        json_like: Mapping[str, JSONed] | Sequence[Mapping[str, JSONed]],
        shared_data: Mapping[str, ObjectList[JSONable]],
    ) -> Self:
        def from_json_like_item(
            child_spec: ChildObjectSpec, json_like_i: JSONed
        ) -> JSONable:
            if not (
                child_spec.class_name
                or child_spec.class_obj
                or child_spec.is_multiple
                or child_spec.shared_data_name
            ):
                # Nothing to process:
                return cast("JSONable", json_like_i)

            # (if is_dict_values) indices into multi_chd_objs that enable reconstruction
            # of the source dict:
            multi_chd_objs, is_dict_values_idx = cls.__remap_child_seq(
                child_spec, json_like_i
            )

            out: list[JSONable] = []
            if child_spec.shared_data_name:
                for i in multi_chd_objs:
                    if i is None:
                        out.append(i)
                        continue

                    sd_lookup_kwargs: dict[str, JSONable]
                    if isinstance(i, str):
                        if i.startswith("hash:"):
                            sd_lookup_kwargs = {"_hash_value": i.removeprefix("hash:")}
                        else:
                            assert child_spec.shared_data_primary_key
                            sd_lookup_kwargs = {child_spec.shared_data_primary_key: i}
                    elif isinstance(i, dict):
                        sd_lookup_kwargs = i
                    else:
                        raise TypeError(
                            "Shared data reference must be a str or a dict."
                        )  # TODO: test raise
                    out.append(
                        shared_data[child_spec.shared_data_name].get(**sd_lookup_kwargs)
                    )
            else:
                chd_cls = cls.__get_child_class(child_spec)
                assert chd_cls is not None
                if issubclass(chd_cls, enum.Enum):
                    out = cls.__inflate_enum(chd_cls, multi_chd_objs)
                else:
                    out.extend(
                        (
                            None
                            if item is None
                            else chd_cls.from_json_like(
                                cast("Any", item),  # FIXME: This is "Trust me, bro!" hack
                                shared_data,
                            )
                        )
                        for item in multi_chd_objs
                    )

            if child_spec.is_dict_values:
                if child_spec.is_dict_values_ensure_list:
                    return {k: [out[i] for i in v] for k, v in is_dict_values_idx.items()}
                else:
                    return {k: out[v[0]] for k, v in is_dict_values_idx.items()}

            elif not child_spec.is_multiple:
                return out[0]

            return out

        if cls._validation_schema:
            validation_schema = get_schema(cls._validation_schema)
            validated = validation_schema.validate(json_like)
            if not validated.is_valid:
                raise ValueError(validated.get_failures_string())

        json_like_copy = copy.deepcopy(json_like)

        for child_spec in cls._child_objects:
            if child_spec.is_single_attribute:
                if len(cls._child_objects) > 1:
                    raise TypeError(
                        f"If ChildObjectSpec has `is_single_attribute=True`, only one "
                        f"ChildObjectSpec may be specified on the class. Specified child "
                        f"objects specs are: {cls._child_objects!r}."
                    )
                json_like_copy = {child_spec.name: json_like_copy}

            assert isinstance(json_like_copy, Mapping)
            if child_spec.json_like_name and child_spec.json_like_name in json_like_copy:
                json_like_copy = dict(json_like_copy)
                json_like_copy[child_spec.name] = cast(
                    "JSONed",
                    from_json_like_item(
                        child_spec, json_like_copy.pop(child_spec.json_like_name)
                    ),
                )

        assert isinstance(json_like_copy, Mapping)

        need_hash = hasattr(cls, "_hash_value") and "_hash_value" not in json_like_copy

        try:
            if issubclass(cls, _AltConstructFromJson):
                obj = cls._json_like_constructor(json_like_copy)
            else:
                obj = cls(**json_like_copy)
        except TypeError as err:
            raise TypeError(
                f"Failed initialisation of class {cls.__name__!r}. Check the signature. "
                f"Caught TypeError: {err}"
            ) from err

        if need_hash:
            obj._set_hash()
        return obj

    def __set_parent_ref(self, chd_obj: Any, child_spec: ChildObjectSpec):
        if chd_obj is not None:
            assert child_spec.parent_ref
            setattr(chd_obj, child_spec.parent_ref, self)

    def _set_parent_refs(self, child_name_attrs: Mapping[str, str] | None = None):
        """Assign references to self on child objects that declare a parent ref
        attribute."""
        child_name_attrs = child_name_attrs or {}
        for child_spec in self._child_objects:
            if child_spec.parent_ref:
                chd_name = child_name_attrs.get(child_spec.name, child_spec.name)
                if child_spec.is_multiple:
                    for chd_obj in getattr(self, chd_name):
                        self.__set_parent_ref(chd_obj, child_spec)
                else:
                    self.__set_parent_ref(getattr(self, chd_name), child_spec)

    def _get_hash(self) -> str:
        json_like = self.to_json_like()[0]
        hash_val = self._get_hash_from_json_like(json_like)
        return hash_val

    def _set_hash(self) -> None:
        self._hash_value = self._get_hash()

    @staticmethod
    def _get_hash_from_json_like(json_like) -> str:
        json_like = copy.deepcopy(json_like)
        json_like.pop("_hash_value", None)
        return get_md5_hash(json_like)

    @final
    def to_dict(self) -> dict[str, Any]:
        """
        Serialize this object as a dictionary.
        """
        if hasattr(self, "__dict__"):
            return self._postprocess_to_dict(dict(self.__dict__))
        elif hasattr(self, "__slots__"):
            return self._postprocess_to_dict(
                {var_name: getattr(self, var_name) for var_name in self.__slots__}
            )
        else:
            return self._postprocess_to_dict({})

    def _postprocess_to_dict(self, dct: dict[str, Any]) -> dict[str, Any]:
        """
        Apply any desired postprocessing to the results of :meth:`to_dict`.
        """
        return dct

    def to_json_like(
        self,
        dct: dict[str, JSONable] | None = None,
        shared_data: _JSONDeserState = None,
        exclude: Container[str | None] = (),
        path: list | None = None,
    ) -> tuple[JSONDocument, _JSONDeserState]:
        """
        Serialize this object as an object structure that can be trivially converted
        to JSON. Note that YAML can also be produced from the result of this method;
        it just requires a different final serialization step.
        """
        if dct is None:
            dct_value = {k: v for k, v in self.to_dict().items() if k not in exclude}
        else:
            dct_value = dct

        parent_refs: dict[str, str] = {}
        if self._child_objects:
            for child_spec in self._child_objects:
                if child_spec.is_single_attribute:
                    if len(self._child_objects) > 1:
                        raise TypeError(
                            "If ChildObjectSpec has `is_single_attribute=True`, only one "
                            "ChildObjectSpec may be specified on the class."
                        )
                    assert child_spec.json_like_name is not None
                    dct_value = dct_value[child_spec.json_like_name]

                if child_spec.parent_ref:
                    parent_refs[child_spec.name] = child_spec.parent_ref

        json_like_, shared_data = to_json_like(
            dct_value, shared_data=shared_data, parent_refs=parent_refs, path=path
        )
        json_like: dict[str, JSONed] | list[JSONed] = cast("Any", json_like_)
        shared_data = shared_data or {}

        for child_spec in self._child_objects:
            assert child_spec.json_like_name is not None
            if child_spec.name in json_like:
                assert isinstance(json_like, dict)
                json_like[child_spec.json_like_name] = json_like.pop(child_spec.name)

            if child_spec.shared_data_name:
                assert isinstance(json_like, dict)
                if child_spec.shared_data_name not in shared_data:
                    shared_data[child_spec.shared_data_name] = {}

                chd_obj_js = json_like.pop(child_spec.json_like_name)

                if not child_spec.is_multiple:
                    chd_obj_js = [chd_obj_js]

                shared_keys: list[JSONed] = []
                assert isinstance(chd_obj_js, (list, tuple, set))
                for i in chd_obj_js:
                    if i is None:
                        continue
                    i.pop("_hash_value", None)
                    hash_i = self._get_hash_from_json_like(i)
                    shared_keys.append(f"hash:{hash_i}")
                    shared_data[child_spec.shared_data_name].setdefault(hash_i, i)

                if not child_spec.is_multiple:
                    try:
                        json_like[child_spec.json_like_name] = shared_keys[0]
                    except IndexError:
                        json_like[child_spec.json_like_name] = None
                else:
                    json_like[child_spec.json_like_name] = shared_keys

        return self._postprocess_to_json(json_like), shared_data

    def _postprocess_to_json(self, json_like: JSONDocument) -> JSONDocument:
        return json_like


@hydrate
class JSONLike(BaseJSONLike, AppAware):
    """BaseJSONLike, where the class namespace is the App instance."""

    __sdk_classes: ClassVar[list[type[BaseJSONLike]]] = []

    @classmethod
    def _class_namespace(cls) -> BaseApp:
        return getattr(cls, cls._app_attr)

    @classmethod
    def __get_classes(cls) -> list[type[BaseJSONLike]]:
        """
        Get the collection of actual SDK classes that conform to BaseJSONLike.
        """
        if not cls.__sdk_classes:
            for cls_name in app.sdk_classes:
                cls2 = getattr(app, cls_name)
                if isinstance(cls2, type) and issubclass(cls2, BaseJSONLike):
                    cls.__sdk_classes.append(cls2)
        return cls.__sdk_classes

    @override
    def _postprocess_to_dict(self, d: dict[str, Any]) -> dict[str, Any]:
        out = super()._postprocess_to_dict(d)

        # remove parent references:
        for cls in self.__get_classes():
            for child_spec in cls._child_objects:
                if child_spec.parent_ref:
                    # _SDK_logger.debug(
                    #     f"removing parent reference {chd.parent_ref!r} from child "
                    #     f"object {chd!r}."
                    # )
                    if (
                        self.__class__.__name__ == child_spec.class_name
                        or self.__class__ is child_spec.class_obj
                    ):
                        out.pop(child_spec.parent_ref, None)
        return out
